;==========================================================
;Header
;==========================================================
        .nolist
#ifdef  VTI
    #define TI83P
#endif
#include        "ion.inc"
        .list

#ifdef  TI83P    
        .org    progstart-2
        .db     $BB,$6D
#else
        .org    progstart
#endif
        ret
        jr      nc,start
        .db     "Antrun",0

; Antrun, based on the PC game with the same name. 
; Fun Facts: Has only aligned sprite routines. Uses a 
; 10x6 tilemap. The screenbuffer has switched columns and rows.
; Which means if you LD HL,gbuf and INC HL, you move down a row.
; Same thing with the map: if DE has the map, INC DE moves
; down a tile and ADD DE,[6] moves one tile to the right. Why? 
; Because when you're using only aligned sprite routines 
; it's faster that way! :)
; It also uses some selfmodifying code.

; We only use the middle of the screen, the top and bottom 8 rows
; aren't used, and neither are the left and right columns.
; So the "map" is only 6*10 bytes in size, and the screen
; is only 6*8*10 = 480 bytes.
;==========================================================
; Misc variables
;==========================================================
Map   = gbuf+480    ;There is still space left in the gbuf. Let's use it!
randData = Map+60

;X and Y position of the cursor.
X   = saferam3
Y   = X+1


;The ant's screen coordinates. Note that the upper four
;bits equal the tile the ant is on
Xant = Y+1
Yant = Xant+1

Antdirection  = Yant+1
Borderscrossed = Antdirection+1
;==========================================================
; start of program
;==========================================================
start:
    LD HL,gbuf
    LD De,gbuf+1
    LD BC,767
    LD (HL),$FF
    LDIR
    call IonFastCopy
    LD HL,0                 ;This is one case where I
    LD (X),HL               ;rather NOT have automatic
    LD (score),HL
    LD HL,$0800             ;writeback :-/
    LD (Xant),HL
    LD HL,%0000000001000100
    LD (Antdirection),HL
    DI

;==========================================================
; Initialize the mapdata. Randomly generated
;==========================================================
    CALL InitMapper
    
    LD A,%01010101  ;Just to make sure the ant starts at a
    LD (Map),A      ;legal location
    
;==========================================================
; Tilemapper
;==========================================================
;Considering mapdata:
Tilemapper
    LD BC,Map
    LD DE,gbuf
    LD A,0

TileLoop:
    LD (saferam1),A     ;(saferam1) will be used as a counter
    
    LD A,(BC)
    INC BC

; Note to self, tilessprites in order of appearance:
;Black tile      %11111111
;Cross           %00000000
;Top-Left        %11001100
;Top-Right       %10011001
;Right-Down      %00110011
;Down-Left       %01100011
;Horizontal Line %01010101
;Vertical Line   %10101010
; The reason why I'm not simply increasing the number is because it
; makes changing the orientation of the ant a lot easier later on,
; and hitdetection too.
    PUSH BC
    LD BC,8
    LD HL,Tiles
    INC A \ JR Z,PutTile \ ADD HL,BC \ DEC A
    OR A \ JR Z,PutTile \ ADD HL,BC         ;Yes, this is ugly spaghetti code.
    CP %11001100 \ JR Z,PutTile \ ADD HL,BC ;However, using this method instead
    CP %10011001 \ JR Z,PutTile \ ADD HL,BC ;of neatly increasing the tilenumber
    CP %00110011 \ JR Z,PutTile \ ADD HL,BC ;0,1,2,etc is actually very useful 
    CP %01100110 \ JR Z,PutTile \ ADD HL,BC ;for other parts of the program.
    CP %01010101 \ JR Z,PutTile \ ADD HL,BC ;So useful I consider it worth it.
PutTile:
    LDIR            ;I told you switching columns and rows makes things easier :)

    POP BC
    LD A,(saferam1)
    INC A
    CP 60
    JP NZ,TileLoop

;==========================================================
; Move Cursor
;==========================================================
TestMovement:
    LD A,$FF
    OUT (1),A
    DEC A
    OUT (1),A
    IN A,(1)

    LD B,A    

    LD HL,(X)   ;H = Y, L = X
TestDown:  
    SRL B
    JR C,TestLeft

    INC H
    LD A,6
    SUB H
    JR NZ,TestLeft
    LD H,A
TestLeft:
    SRL B
    JR C,TestRight

    DEC L
    LD A,L
    INC A
    JR NZ,TestRight
    LD L,9
TestRight:
    SRL B
    JR C,TestUp
    
    INC L
    LD A,10
    SUB L
    JR NZ,TestUp
    LD L,A
TestUp:
    SRL B
    JR C,WritebackXY
    
    DEC H
    LD A,H
    INC A
    JR NZ,WritebackXY
    LD H,5

WritebackXY:
    LD (X),HL
;==========================================================
; Check if 2nd was pressed. If so, change tile orientation
; Ceck if Mode was pressed. If so, quit
;==========================================================
    EX DE,HL
;Reminder:  D = Y of cursor
;           E = X of cursor    
Test2nd:
    LD A,$FF
    OUT (1),A
    LD A,%10111111
    OUT (1),A
    IN A,(1)
    CP %11011111
    JR NZ,TestMode

;Test if the cursor is on the same tile as the ant.
;If it is, we can't rotate
    LD HL,(Xant)
    SRL H \ SRL H \ SRL H \ SRL H
    LD A,H
    CP D
    JR NZ,RotateTile
    SRL L \ SRL L \ SRL L \ SRL L
    LD A,L
    CP E
    JR Z,CantRotate
RotateTile
    Call FindTile
    
    RLC (HL)
    JR DontQuit
; To rotate a tile we need only to rotate its byte.
; This is one part where the strange tilenumbering pays off; otherwise we'd been
; forced to use some ugly code (likely worse than in the tilemapper) right now.    

TestMode:
    CP %10111111
    JR NZ,DontQuit
Quit:
    Call BigDelay
    bcall(_ClrLCDFull)
    Call BigDelay
    bcall(_HomeUp)
    LD HL,HighScore
    bcall(_PutS)
    LD HL,$0001
    LD (CurRow),HL
score = $+1
    LD HL,$0000
    bcall(_DispHL)  ;Display Score
    LD HL,18*256
    LD (PenCol),HL
    LD HL,Credit1
    bcall(_VPutS)
    LD DE,24*256
    LD (PenCol),DE
    bcall(_VPutS)
    Call BigDelay
WaitKey
    bcall(_GetCSC)  ;wait for a keypress before moving on
    OR A
    JR Z,WaitKey
    RET
DontQuit
CantRotate   
;==========================================================
; Move the ant. Test hitdetection and change orientation
; if necessary.
;==========================================================    
;Antdirection:
;%10001000 = down
;%01000100 = right
;%00100010 = up
;%00010001 = left
;Upper four bits of Xant and Yant equals the tile it is on,
;and the lower four represent what bit of the tile it is on.
    Call SubMove
    EX DE,HL        ;(DE) = tile ant is on
    LD BC,(Xant)    ;B = (Yant), C = (Xant)
    LD HL,(Score)
    LD A,(Antdirection)
    RLA
    JR C,DownAnt
    RLA
    JR C,RightAnt
    RLA
    JR C,UpAnt

LeftAnt:
    DEC C
    LD A,$0F
    XOR C
    AND $0F         ;Has the ant stepped from one tile on the other?
    JR NZ,OrientationChange

    INC HL          ;Then increase the score by one
    LD (Score),HL

    LD A,(DE)
    OR A            ;Check if the previous tile was a cross
    JR Z,JP1        ;If so, don't change it.
    LD A,$FF        ;else, make it black
    LD (DE),A
JP1    
    SUB C
    JP NZ,HitDetection
    LD HL,Borderscrossed
    INC (HL)
    LD (Bordercross),A
    LD C,9*16+15
    JP HitDetection

DownAnt:
    INC B
    LD A,$0F
    AND B
    JR NZ,OrientationChange
    
    INC HL
    LD (Score),HL
    
    LD A,(DE)
    OR A
    JR Z,JP2
    LD A,$FF
    LD (DE),A
JP2
    LD A,6*16
    Sub B
    JR NZ,HitDetection
    LD HL,Borderscrossed
    INC (HL)    
    LD B,A
    LD (Bordercross),A
    JR HitDetection

RightAnt:
    INC C
    LD A,$0F
    AND C
    JR NZ,OrientationChange
    
    INC HL
    LD (Score),HL
    
    LD A,(DE)
    OR A
    JR Z,JP3
    LD A,$FF
    LD (DE),A
JP3
    LD A,10*16
    Sub C
    JR NZ,HitDetection
    LD HL,Borderscrossed
    INC (HL)    
    LD C,A
    LD (Bordercross),A
    JR HitDetection

UpAnt:
    DEC B
    LD A,$0F
    XOR B
    AND $0F
    JR NZ,OrientationChange

    INC HL
    LD (Score),HL
    
    LD A,(DE)
    OR A
    JR Z,JP4
    LD A,$FF
    LD (DE),A
JP4
    sub B
    JR NZ,HitDetection
    LD HL,Borderscrossed
    INC (HL)
    LD (Bordercross),A
    
    LD B,5*16+15
    JR HitDetection

    
; Another benefit of the strange tilenumbering: It makes
; changing the ant's direction easy. Why? take a good look:
;Antdirection:
;%10001000 = down
;%01000100 = right
;%00100010 = up
;%00010001 = left
;Tiledata:
;Black tile      %11111111
;Cross           %00000000
;Top-Left        %11001100
;Top-Right       %10011001
;Right-Down      %00110011
;Down-Left       %01100011
;Horizontal Line %01010101
;Vertical Line   %10101010

OrientationChange:
    LD (Xant),BC    ;write back the modified coordinates
    LD A,%00001111
    AND B
    CP $08
    JR NZ,EndMovement
    LD A,%00001111
    AND C
    CP $08    ;If the lower four bits of H and L are both equal to 8
    JR NZ,EndMovement   ;the ant is in the middle of the tile. Then and only
    Call SubMove        ;then do we change the orientation.
    LD A,(HL)
    OR A
    JR Z,EndMovement
    LD A,(AntDirection)
    XOR (HL)            ;(HL) = tile ant is on.
    RLCA \ RLCA
    LD (AntDirection),A
    JR EndMovement

; And the final benefit of the tilenumbering: 
; It eases up hitdetection... a lot.
HitDetection:
    LD (Xant),BC    ;write back the modified coordinates
    Call SubMove
    LD A,(HL)
    OR A            ;is it a cross?
    JR Z,EndMovement
    INC A           ;is it a black tile?
    JP Z,Quit
    LD A,(AntDirection)
    AND (HL)
    JP Z,Quit

EndMovement
;==========================================================
; Write the ant on the tile in gbuf.
; Write the cursor on the tile in gbuf.
;==========================================================
OutputAnt:
    LD DE,(Xant)    ;D = Yant, E = Xant
    SRL D
    LD C,D
    LD B,0
    LD HL,gbuf
    ADD HL,BC
    LD A,%11110000
    AND E
    LD C,A
    ADD HL,BC
    ADD HL,BC
    ADD HL,BC

    LD A,E
    RRA
    AND %00000111
    INC A
    LD B,A
    LD A,1
OutputAntLoop:
    RRCA
    DJNZ OutputAntLoop
    XOR (HL)
    LD (HL),A
    
OutputCursor:
    LD HL,gbuf+1
    LD A,(Y)
    ADD A,A         ;2*y
    ADD A,A         ;4*y
    ADD A,A         ;8*Y
    LD C,A
    ADD HL,BC
    LD A,(X)
    ADD A,A         ;2*X
    ADD A,A         ;4*X
    ADD A,A         ;8*X
    LD C,A    
    ADD A,A         ;16*X
    LD C,A
    ADD HL,BC
    ADD HL,BC
    ADD HL,BC       ;48*X
    LD B,7
OutputCursorLoop:
    LD A,%01111111  ;The cursor is nothing but an inverse
    XOR (HL)        ;of whatever tile it is on
    LD (HL),A
    INC HL
    djnz OutputCursorLoop    

;==========================================================
; Copy the buffer to the screen
;==========================================================
    LD C,$21
    LD HL,gbuf
Screencopyloop1:
    NOP                     ;4
    NOP                     ;4
    LD A,C                  ;4
    OUT ($10),A             ;11
    CALL LCDBusy            ;enough
    LD B,48
    LD A,$88
    OUT ($10),A             ;11
Screencopyloop2:
    LD A,(HL)               ;7
    Call LCDBusy            ;enough
    OUT ($11),A             ;11
    INC HL                  ;6
    djnz Screencopyloop2    ;13/8
    INC C                   ;4
    LD A,$2B                ;7
    Sub C                   ;4
    JR NZ,Screencopyloop1   ;13
    Call Delay


;==========================================================
; Checking if a border was crossed, and if so reset the
; black tiles. Change the chance of spawning a bad tile.
;==========================================================
    LD A,(Borderscrossed)
    RRA
    RRA
    RRA
    AND %00011111
    LD B,A
    LD HL,(Score)
    LD A,H
    RL D
    RLA
    JR C,MaxProb
    RL D
    RLA
    JR C,MaxProb
    RL D
    RLA
    JR C,MaxProb
    ADD A,B
    JR C,MaxProb        ;If there was carry, the chance of
    ADD A,16            ;a black tile is even higher than
    JR C,MaxProb        ;the maximum the random routine can
    LD (randomSMC),A    ;handle, so we just maximize it.
    JR Bordercross-1
MaxProb
    XOR A               ;The maximum chance is equal to running DJNZ 256
    LD (randomSMC),A    ;times, which equals loading 0 into B
    
;The chances of getting a black tile are equal to A+B/(16+A+B),
;where A is the score divided by 32, rounded down, and B is
;the number of times you crossed the border divided by 8,
;rounded down

Bordercross = $+1           ;If the ant crossed the border,
    LD A,1                  ;we reset all the black tiles with
    OR A                    ;the InitMapper part of the code.
    LD A,1                  ;Though we do want to make sure this
    LD (Bordercross),A      ;doesn't happen each loop.
    CALL Z, InitMapper
    JP Tilemapper
;==========================================================
;==========================================================
; Misc subroutines
;==========================================================
InitMapper:
    LD B,60
    LD HL,Map

InitLoop:
    LD A,(HL)
    INC A
    JR NZ,SkipTile
    Call randomNR
    LD C,%01010101
    sub 4 \ JR C,RandomOrientation  ;It's a line
    LD C,%11001100
    sub 12 \ JR C,RandomOrientation ;It's a hook
    LD C,$FF                        ;It's a black tile.
;Initially, you can't get a black tile, but as your score
;the chances of one appearing increase
RandomOrientation:
    call randomNR
    INC A
    PUSH BC
    LD B,A
    LD A,C
OrientationLoop:    ;randomize the orientation.
    RLCA
    DJNZ OrientationLoop
    POP BC
    LD (HL),A
SkipTile:
    INC HL
    DJNZ InitLoop
    XOR A
    LD (Map+13),A   ;set the four crosses
    LD (Map+22),A
    LD (Map+40),A
    LD (Map+49),A
    RET

;Generate a random number between 0-15, initially
;In a later version the program will use SMC to modify the
;outputs, for creating the chance of getting a 'bad' tile
; ouput a = answer, 0 <= a <= 15 initially
; destroys AF
randomNR
    PUSH BC
    PUSH HL
    PUSH DE
    ld b,16
randomSMC = $-1
    ld  hl,(randData)
    ld  a,r
    ld  d,a
    ld  e,(hl)
    add hl,de
    add a,l
    xor h
    ld  (randData),hl
    sbc hl,hl
    ld  e,a
    ld  d,h
randomLoop:
    add hl,de
    djnz    randomLoop
    ld  a,h
    POP DE
    POP HL
    POP BC
    ret
    
LCDBusy
    PUSH HL
    INC DE
    DEC DE
    POP HL
    RET

SubMove:
;Input: None
;Output: (HL) = underlying tile
;Destroys all
    LD DE,(Xant)
    LD HL,Map
    LD A,E
    LD B,0
    AND $F0 ;A = 16*X (tiles)
    RRA     ;8*X
    RRA     ;4*X
    LD C,A
    RRA
    ADD A,C ;6*X
    LD C,A
    ADD HL,BC
    LD A,D
    AND $F0
    RRA
    RRA
    RRA
    RRA
    LD C,A
    ADD HL,BC
    RET
    
FindTile
;Input: D = Y, E = X
;Output: (HL)=underlying tile
;Destroys: All but IX,IY
    LD B,0
    LD C,D
    LD HL,Map

    ADD HL,BC   ;Add Map,Y

    LD A,E
    ADD A,A     ; 2 * X
    LD E,A    
    ADD A,A     ; 4 * X
    ADD A,E     ; 6 * X
    LD C,A
    ADD HL,BC   ;Add Map,6*X
    RET

BigDelay:
    LD C,16
BigDelayLoop
    Call Delay
    DEC C
    JR NZ,BigDelayLoop
    RET

Delay:
    XOR A
    LD B,A
DelayLoop:                  ;This program is sickeningly fast and it isn't even 
    DJNZ DelayLoop          ;fully optimized I probably should put in some more 
    DEC A                   ;size optimizations because speed is not an issue.
    JR NZ,DelayLoop
    RET
;==========================================================
; Tiledata
;==========================================================

Tiles:
 .db $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF ;Black tile      %11111111
 .db $FF,$E3,$E3,$80,$80,$80,$E3,$E3 ;Cross           %00000000
 .db $FF,$E3,$E3,$83,$83,$83,$FF,$FF ;Top-Left Hook   %11001100
 .db $FF,$E3,$E3,$E0,$E0,$E0,$FF,$FF ;Top-Right Hook  %10011001
 .db $FF,$FF,$FF,$E0,$E0,$E0,$E3,$E3 ;Right-Down Hook %00110011
 .db $FF,$FF,$FF,$83,$83,$83,$E3,$E3 ;Down-Left Hook  %01100011
 .db $FF,$FF,$FF,$80,$80,$80,$FF,$FF ;Horizontal Line %01010101
 .db $FF,$E3,$E3,$E3,$E3,$E3,$E3,$E3 ;Vertical Line   %10101010
HighScore:
    .db "Your Score Was:",0
Credit1:
    .db "Ver 1.00 By Job, Feb 2004",0
Credit2:
    .db "Greenfire Productions",0
.end
.END